package com.hisschemoller.audio.model 
{
	import com.hisschemoller.audio.model.vo.SoundVO;
	import com.hisschemoller.audio.utils.delegates.AudioDecoderDelegate;
	import com.hisschemoller.audio.utils.delegates.AudioLoaderDelegate;
	import com.hisschemoller.audio.utils.delegates.ByteArrayToVectorDelegate;
	import com.hisschemoller.audio.utils.delegates.vo.AudioLoaderVO;
	import com.hisschemoller.ui.model.ICanBeBusy;
	import com.hisschemoller.ui.model.vo.AlertVO;
	import com.hisschemoller.ui.notification.UINotifications;
	import com.hisschemoller.utils.responder.Responder;

	import org.puremvc.as3.multicore.interfaces.IProxy;
	import org.puremvc.as3.multicore.patterns.proxy.Proxy;

	import flash.utils.getQualifiedClassName;

	/**
	 * AudioLibraryProxy loads and manages audio files.
	 * Each audio file is loaded once and is stored in the _sounds vector.
	 * When an instrument requests a sound to play it receives a copy of the
	 * SoundVO object. So the sounds in the library are the master versions that
	 * are never actually used. This is done so the instruments can perform
	 * any destructive editing on the the sound they want, as there will still 
	 * be a pristine copy in the library.
	 * 
	 * @author Wouter Hisschemoller
	 * (c) Feb 19, 2010
	 */
	public class AudioLibraryProxy extends Proxy implements IProxy, ICanBeBusy
	{
		public static const NAME : String = getQualifiedClassName( AudioLibraryProxy );
		private var _sounds : Vector.<SoundVO> = new Vector.<SoundVO>( );
		private var _isBusy : Boolean;

		public function AudioLibraryProxy ()
		{
			super( NAME );
		}

		/**
		 * Add a new sound to the library.
		 */
		public function addSound (soundVO : SoundVO) : void
		{
			/** Don't add if the sound already exists. */
			if(soundExists( soundVO )) return;
			
			/** Add the sound. */
			_sounds.push( soundVO );
		}

		/**
		 * When an instrument wants to play a sound it gets a copy of the requested sound.
		 * This leaves a clean copy in the library while the instrument might perform
		 * destructive editing on the sound.
		 */
		public function getCopyOfSoundByID ( id : String ) : SoundVO
		{
			var n : int = _sounds.length;
			while(--n > -1)
			{
				if(_sounds[n].id == id) return _sounds[n].clone( );
			}
			
			return null;
		}

		/**
		 * Get reference to a SoundVO object. 
		 * Do not change the properties of the SoundVO. The library copies must remain unchanged.
		 * Use this if you need to retrieve data from the SoundVO.
		 * Used by: SamplerInstrumentProxy.addProgram(). Probably temporarily.
		 */
		public function getReferenceToSoundByID ( id : String ) : SoundVO
		{
			var n : int = _sounds.length;
			while(--n > -1)
			{
				if(_sounds[n].id == id) return _sounds[n];
			}
			
			return null;
		}

		/**
		 * Load all audioFiles in the vector.
		 */
		public function loadAudioFiles (sounds : Vector.<AudioLoaderVO>) : void
		{
			/** Skip sounds that are already in the soundbank. */
			var n : int = sounds.length;
			while(--n > -1)
			{
				var m : int = _sounds.length;
				for(var j : uint = 0; j < m; j++)
				{
					if(sounds[n].url == _sounds[j].url)
					{
						sounds.splice( n, 1 );
						break;
					}
				}
			}
			
			if(sounds.length > 0)
			{
				setBusy( true );
				
				new AudioLoaderDelegate( new Responder( loadResult, fault ) ).loadAudioFiles( sounds );
			}
		}

		/**
		 * 
		 */
		public function get isBusy () : Boolean
		{
			return _isBusy;
		}
    
		/**
		 * 
		 */
		private function setBusy (isBusy : Boolean) : void
		{
			_isBusy = isBusy;
			
			facade.sendNotification( UINotifications.SET_BUSY, _isBusy );
		}

		/**
		 * Audio files succesfully loaded.
		 */
		private function loadResult (loadedStack : Vector.<AudioLoaderVO>) : void
		{
			new AudioDecoderDelegate( new Responder( decodeResult, fault ) ).decodeAudio( loadedStack );
		}

		/**
		 * An error occurred while loading / decoding a sound.
		 */
		private function fault (info : String) : void
		{
			log( "e A load / decode error happened:  " + info );
			
			setBusy( false );
		}

		/**
		 * Audio files succesfully decoded.
		 */
		private function decodeResult (decodedStack : Vector.<AudioLoaderVO>) : void
		{
			new ByteArrayToVectorDelegate( new Responder( convertResult, convertFault ) ).convertAudio( decodedStack, AudioDriverProxy.BYTES_PER_MONO_SAMPLE );
		}

		/**
		 * Audio file failed to open.
		 */
		private function convertFault (info : String) : void
		{
			facade.sendNotification( UINotifications.ALERT, new AlertVO( "File convert error", info ) );
			
			setBusy( false );
		}

		/**
		 * Audio converted to Vector. 
		 */
		private function convertResult (convertedStack : Vector.<AudioLoaderVO>) : void
		{
			var n : int = convertedStack.length;
			while(--n > -1)
			{
				var audioLoaderVO : AudioLoaderVO = convertedStack[n];
				
				var soundVO : SoundVO = new SoundVO( );
				soundVO.id = audioLoaderVO.id;
				soundVO.url = audioLoaderVO.url;
				soundVO.byteArray = audioLoaderVO.byteArray;
				soundVO.channels = audioLoaderVO.channels;
				soundVO.samples = audioLoaderVO.samples;
				soundVO.numSamplesTotal = audioLoaderVO.numSamplesTotal;
				_sounds.push( soundVO );
			}
			
			setBusy( false );
		}

		/**
		 * Checks if a SoundVO exists.
		 */
		private function soundExists (soundVO : SoundVO) : Boolean
		{
			var n : int = _sounds.length;
			while(--n > -1)
			{
				if(soundVO == _sounds[n]) return true;
			}
			
			return false;
		}
	}
}