package com.hisschemoller.sequencer.view.components.linearpattern
{
	import com.bit101.components.HSlider;
	import com.bit101.components.Panel;
	import com.hisschemoller.sequencer.model.enum.PlaybackStatus;
	import com.hisschemoller.sequencer.model.vo.PlaybackVO;
	import com.hisschemoller.sequencer.model.vo.SeqEventVO;
	import com.hisschemoller.sequencer.model.vo.SequenceVO;
	import com.hisschemoller.sequencer.model.vo.TempoVO;
	import com.hisschemoller.sequencer.view.enum.UILayout;
	import com.hisschemoller.utils.IDisposable;

	import flash.display.DisplayObjectContainer;
	/**
	 * @author wouter.hisschemoller
	 * Oct 19, 2010
	 */
	public class PatternEditor extends Panel implements IDisposable
	{
		private static const TRACKS_Y : int = 20;
		private static const TRACK_HEIGHT : int = 22;
		private static const MARGIN_RIGHT : int = 20;
		private var _positionBar : HSlider;
		private var _tracks : Vector.<PatternEditorTrack> = new Vector.<PatternEditorTrack>( );
		private var _zoomLevel : Number = 1;
		private var _sequenceLength : int = 1;
		private var _millisecondsPerPart : Number;
		private var _positionInPartsAtProcessStart : Number;
		private var _timeAtProcessStart : Number;

		public function PatternEditor ( parent : DisplayObjectContainer = null, xpos : Number = 0, ypos : Number = 0 )
		{
			super( parent, xpos, ypos );
		}

		/**
		 * Clean up before when intance will be deleted.
		 */
		public function dispose () : void
		{
			var n : int = _tracks.length;
			while(--n > -1)
			{
				_tracks[n].dispose( );
			}
			
			_tracks = null;
		}

		/**
		 * Update the time slider position (on ENTER_FRAME).
		 */
		public function updateTime (latencyInMs : Number) : void
		{
			/** Number of parts since the last SAMPLE_DATA event. */
			var offsetInParts : Number = ( new Date( ).getTime( ) - _timeAtProcessStart ) / _millisecondsPerPart;
			
			/** Latency in parts. */
			var latencyInParts : Number = latencyInMs / _millisecondsPerPart;
			
			_positionBar.value = ( _positionInPartsAtProcessStart + offsetInParts - latencyInParts ) % _sequenceLength;
		}

		/**
		 * On SeqNotifications.SEQ_EVENTS_PROCESSED.
		 */
		public function updatePlayback (playbackVO : PlaybackVO) : void
		{
			_positionInPartsAtProcessStart = playbackVO.positionInPartsAtProcessStart;
			_timeAtProcessStart = playbackVO.timeAtProcessStart;
		}

		/**
		 * Tempo has changed. 
		 * _millisecondsPerPart is used in the time indicator animation.
		 */
		public function updateTempo (tempoVO : TempoVO) : void
		{
			_millisecondsPerPart = tempoVO.millisecondsPerPart;
		}

		/**
		 * 
		 */
		public function updateZoomLevel (zoomLevel : Number) : void
		{
			_zoomLevel = zoomLevel;
			
			var n : uint = _tracks.length;
			for( var i : uint = 0; i < n; i++ )
			{
				_tracks[i].pattern.updateZoomLevel( _zoomLevel );
			}
		}

		/**
		 * Playback started or stopped.
		 */
		public function updatePlaybackState (playbackStatus : String) : void
		{
			if(playbackStatus != PlaybackStatus.PLAYING)
			{
				/** Reset play bar  */
				_positionBar.value = 0;
			}
		}

		/**
		 * Add a new track to the editor.
		 */
		public function addTrack ( trackID : String ) : void
		{
			var track : PatternEditorTrack = new PatternEditorTrack( trackID, this, 0, 0 );
			addChild( track );

			_tracks.push( track );

			invalidate( );
		}

		/**
		 * Sequence has changed.
		 */
		public function changeSequence ( sequenceVO : SequenceVO ) : void
		{
			_sequenceLength = sequenceVO.length;
			_positionBar.maximum = _sequenceLength;
			var m : uint = sequenceVO.patterns.length;
			var n : uint = _tracks.length;
			for( var i : uint = 0; i < n; i++ )
			{
				var track : PatternEditorTrack = _tracks[ i ];
				for( var j : uint = 0; j < m; j++ )
				{
					if(track.id == sequenceVO.patterns[ j ].trackID )
					{
						track.updatePattern( sequenceVO.patterns[ j ] );
						break;
					}
				}
			}
			
			/** Set the zoomLevel so the sequence fills the avaiable width for the pattern. */
			updateZoomLevel( (stage.stageWidth - PatternEditorTrack.PATTERN_X - MARGIN_RIGHT - (UILayout.MARGIN * 2)) / _sequenceLength );

			invalidate( );
		}

		/**
		 * 
		 */
		public function addEvent (seqEventVO : SeqEventVO) : void
		{
			var n : uint = _tracks.length;
			while(--n > -1)
			{
				var track : PatternEditorTrack = _tracks[n];
				if(track.id == seqEventVO.trackID)
				{
					track.pattern.addStep( seqEventVO );
					return;
				}
			}
		}

		/**
		 * 
		 */
		public function deleteEvent (seqEventVO : SeqEventVO) : void
		{
			var n : uint = _tracks.length;
			while(--n > -1)
			{
				var track : PatternEditorTrack = _tracks[n];
				if(track.id == seqEventVO.trackID)
				{
					track.pattern.deleteStep( seqEventVO );
					return;
				}
			}
		}

		/**
		 * Redraw the whole editor.
		 */
		override public function draw () : void
		{
			super.draw( );

			var n : uint = _tracks.length;
			for( var i : uint = 0; i < n; i++ )
			{
				_tracks[i].y = TRACKS_Y + (i * TRACK_HEIGHT);
			}

			_positionBar.y = TRACKS_Y + (_tracks.length * TRACK_HEIGHT);
			_positionBar.setSize( _zoomLevel * _sequenceLength, _positionBar.height );
		}

		/**
		 * Component init, part of the startup cyclus.
		 */
		override protected function init () : void
		{
			super.init( );
			
			move( UILayout.MARGIN, UILayout.CONTROLS_HEIGHT + UILayout.MARGIN + UILayout.VERTICAL_GAP );
			setSize( stage.stageWidth - ( UILayout.MARGIN * 2 ), stage.stageHeight - UILayout.CONTROLS_HEIGHT - ( UILayout.MARGIN * 2 ) - UILayout.VERTICAL_GAP );
		}

		/**
		 * Add displayObjects, part of the startup cyclus.
		 */
		override protected function addChildren () : void
		{
			super.addChildren( );

			_positionBar = new HSlider( this, PatternEditorTrack.PATTERN_X, TRACKS_Y );
			_positionBar.minimum = 0;
		}
	}
}
