package com.hisschemoller.sequencer.view.components.linearpattern 
{
	import com.hisschemoller.sequencer.model.vo.PatternVO;
	import com.hisschemoller.sequencer.model.vo.SeqEventVO;
	import com.hisschemoller.sequencer.view.events.SequencerEvent;
	import com.hisschemoller.utils.IDisposable;
	import com.newgonzo.midi.messages.MessageStatus;

	import flash.display.DisplayObjectContainer;
	import flash.events.MouseEvent;
	import flash.events.TimerEvent;
	import flash.utils.Timer;
	/**
	 * Step button MOUSE_DOWN:
	 * 1. Bewaar x-positie knop.
	 * 2. Bewaar time van de knop.
	 * 3. Dispatch DELETE_EVENT. (knop wordt verwijderd)
	 * 4. Maak nieuwe tijdelijke knop op dezelfde x-positie.
	 * 5. MOUSE_MOVE listener op tijdelijke knop.
	 * 6. MOUSE_UP handler op tijdelijke knop.
	 * 
	 * Step button MOUSE_UP:
	 * 1. Verhoog _numClicks.
	 * 2. Start de _doubleClickTimer.
	 * 
	 * Step button MOUSE_DOWN (binnen timer tijd):
	 * 1. Doe niks.
	 * 
	 * Step button MOUSE_UP (binnen timer tijd):
	 * 1. Verhoog _numClicks.
	 * 2. Start de _doubleClickTimer.
	 * 
	 * Timer afgelopen bij _numClicks == 1 (CLICK):
	 * 1. Bereken time van de knop (op eventueel nieuwe positie).
	 * 2. Verwijder MOUSE_MOVE listener op tijdelijke knop.
	 * 3. Verwijder MOUSE_UP listener op tijdelijke knop.
	 * 4. Verwijder de tijdelijke knop.
	 * 5. Dispatch CREATE_EVENT. (nieuwe knop wordt gemaakt)
	 * 
	 * Timer afgelopen bij _numClicks != 1 (DOUBLE_CLICK):
	 * 1. Verwijder MOUSE_MOVE listener op tijdelijke knop.
	 * 2. Verwijder MOUSE_UP listener op tijdelijke knop.
	 * 3. Verwijder de tijdelijke knop.
	 * 
	 * @author wouterhisschemoller
	 */
	public class PatternEditorPatternFree extends PatternEditorPatternBase implements IDisposable
	{
		private var _temporaryDragStep : PatternEditorStep;
		private var _lastAddedStep : PatternEditorStep;
		private var _doubleClickTimer : Timer;
		private var _numClicks : uint;

		public function PatternEditorPatternFree (parent : DisplayObjectContainer = null, xpos : Number = 0, ypos : Number = 0, h : Number = 0)
		{
			super( parent, xpos, ypos, h );
		}

		/**
		 * Clean up before deletion.
		 */
		public function dispose () : void
		{
			_background.removeEventListener( MouseEvent.DOUBLE_CLICK, backgroundDoubleClickHandler );
			removeTemporaryStep( );
			removeStageListener( );
			
			_doubleClickTimer.stop( );
			_doubleClickTimer.removeEventListener( TimerEvent.TIMER_COMPLETE, doubleClickTimerHandler );
			_doubleClickTimer = null;
		}

		/**
		 * Add steps to the pattern.
		 */
		override public function updatePattern (patternVO : PatternVO) : void
		{
			super.updatePattern( patternVO );
			
			/** Create steps as nessecary. */
			var numEvents : uint = patternVO.events.length;
			for(var i : uint = 0; i < numEvents; i++ ) 
			{
				var seqEventVO : SeqEventVO = patternVO.events[i];

				/** NOTE_ON events are represented as selected steps. */
				if( seqEventVO.message.status == MessageStatus.NOTE_ON ) 
				{
					var xPos : Number = (seqEventVO.time / patternVO.length) * _width;
					_lastAddedStep = new PatternEditorStep( seqEventVO, this, xPos, 0 );
					_lastAddedStep.setSize( _minimalStepButtonWidth, _height );
					_lastAddedStep.addEventListener( MouseEvent.MOUSE_DOWN, stepMouseDownHandler );
					_steps.push( _lastAddedStep );
				}
			}
		}

		/**
		 * Zoom level changed.
		 */
		override public function updateZoomLevel (zoomLevel : Number) : void
		{
			super.updateZoomLevel( zoomLevel );
			
			/** Update the step positions. */
			var n : int = _steps.length;
			while( --n > -1 )
			{
				var step : PatternEditorStep = _steps[n];
				step.x = ( step.seqEventVO.time / _length ) * ( _length * _zoomLevel );
			}
		}

		/**
		 * Add a new step.
		 */
		override public function addStep (seqEventVO : SeqEventVO) : PatternEditorStep
		{
			_lastAddedStep = super.addStep( seqEventVO );
			_lastAddedStep.x -= _lastAddedStep.width / 2;
			_lastAddedStep.addEventListener( MouseEvent.MOUSE_DOWN, stepMouseDownHandler );
			return _lastAddedStep;
		}

		/**
		 * Component init, part of the startup cyclus.
		 */
		override protected function init () : void
		{
			super.init( );
			
			_doubleClickTimer = new Timer( 250, 1 );
			_doubleClickTimer.addEventListener( TimerEvent.TIMER_COMPLETE, doubleClickTimerHandler );
			
			_background.doubleClickEnabled = true;
			_background.addEventListener( MouseEvent.DOUBLE_CLICK, backgroundDoubleClickHandler );
		}

		/**
		 * Remove the temporary step button that is used for dragging.
		 */
		private function removeTemporaryStep () : void
		{
			if(_temporaryDragStep)
			{
				_temporaryDragStep.dispose( );
				removeChild( _temporaryDragStep );
				_temporaryDragStep = null;
			}
		}

		/**
		 * Double click on background creates a new step.
		 */
		private function backgroundDoubleClickHandler (event : MouseEvent) : void 
		{
			createNewEvent( mouseX );
		}

		/**
		 * Mouse down on one of the step buttons.
		 */
		private function stepMouseDownHandler (event : MouseEvent) : void 
		{
			if(!_doubleClickTimer.running)
			{
				/** Initialize doubleclick sequence. */
				_numClicks = 0;
				
				/** Delete step. */
				var step : PatternEditorStep = event.target as PatternEditorStep;
				dispatchEvent( new SequencerEvent( SequencerEvent.EVENT_DELETE, step.seqEventVO, true ) );
				
				/** Create temporary step to drag. */
				_temporaryDragStep = new PatternEditorStep( null, this, step.x, 0 );
				_temporaryDragStep.setSize( 10, _height );
				_temporaryDragStep.alpha = 0.6;
				
				addStageListener( );
			}
		}

		/**
		 * MouseUp during step doubleclick cyclus.
		 */
		private function stageMouseUpHandler (event : MouseEvent) : void 
		{
			/** First mouseUp, remove temp step and add real one. */
			if(!_doubleClickTimer.running)
			{
				removeTemporaryStep( );
				createNewEvent( mouseX );
			}
			
			/** Increase click count. */
			_numClicks++;
			
			if( _numClicks == 2 )
			{
				/** If this is the second click remove the step. */
				dispatchEvent( new SequencerEvent( SequencerEvent.EVENT_DELETE, _lastAddedStep.seqEventVO, true ) );
			}
			else
			{
				_doubleClickTimer.start( );
			}
		}

		/**
		 * Move the temporary drag button.
		 */
		private function stageMouseMoveHandler (event : MouseEvent) : void 
		{
			if(_temporaryDragStep)
			{
				_temporaryDragStep.x = Math.max( 0, Math.min( mouseX - (_temporaryDragStep.width / 2), _width ) );
			}
		}

		/**
		 * Timer ended. Remove event listeners.
		 */
		private function doubleClickTimerHandler (event : TimerEvent) : void 
		{
			removeStageListener( );
		}

		/**
		 * Add mouse event listeners to stage.
		 */
		private function addStageListener () : void
		{
			stage.addEventListener( MouseEvent.MOUSE_MOVE, stageMouseMoveHandler, true );
			stage.addEventListener( MouseEvent.MOUSE_UP, stageMouseUpHandler, true );
		}

		/**
		 * Remove mouse event listeners from stage.
		 */
		private function removeStageListener () : void
		{
			stage.removeEventListener( MouseEvent.MOUSE_MOVE, stageMouseMoveHandler, true );
			stage.removeEventListener( MouseEvent.MOUSE_UP, stageMouseUpHandler, true );
		}
	}
}
